package au.asn.ucc.ticker;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.DisplayMode;
import java.awt.Graphics;
import java.awt.GraphicsDevice;
import java.awt.Image;
import java.util.ArrayList;
import java.util.Date;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Class that can display something on a TickerFrame. Tickers have a priority
 * which determines when they are activated to be displayed, a Ticker will only
 * be activated if there is no Ticker with higher priority. Once a ticker is
 * activated it will be executing updateLoop from inside the separate thread
 * updateThread, it should not exit from updateLoop until updateThread is
 * interrupted.
 * 
 * $Id: Ticker.java 36 2009-12-30 11:32:21Z rvvs89 $
 * @author rvvs89
 */
public abstract class Ticker implements Comparable<Ticker> {

    private PriorityEnum priority;
    private boolean active;
    private Date lastShown;
    private Image image;
    private GraphicsDevice device;
    private Dimension dimension;
    private final ArrayList<PriorityChangeListener> listeners;
    private final Thread renderThread;
    private final Thread updateThread;
    private final Thread backgroundThread;
    private final String name;
    private final long id;
    private final AtomicBoolean dirty;
    private static final Date INITIAL_LAST_SHOWN = new Date(0L);
    private static final AtomicLong UID = new AtomicLong();

    /**
     * Constructs a new instance of Ticker.
     */
    protected Ticker(final String name) {
        this.priority = PriorityEnum.DISABLED;
        lastShown = INITIAL_LAST_SHOWN;
        active = false;
        image = null;
        device = null;
        dimension = null;
        id = UID.incrementAndGet();
        listeners = new ArrayList<PriorityChangeListener>(1);
        dirty = new AtomicBoolean(false);
        renderThread = new Thread(new Runnable() {

            public void run() {
                synchronized (renderThread) {
                    synchronized (Ticker.this) {
                        Ticker.this.notify();
                    }
                    while (true) {
                        try {
                            renderThread.wait();
                        } catch (InterruptedException ex) {
                            Logger.getLogger(Ticker.class.getName()).log(Level.SEVERE, null, ex);
                            System.exit(1);
                        }
                        image = render();
                        dirty.set(true);
                        renderThread.notify();
                    }
                }
            }
        });
        renderThread.setDaemon(true);
        renderThread.setName("Ticker" + id + "(" + name + "): renderThread");
        updateThread = new Thread(new Runnable() {

            public void run() {
                synchronized (updateThread) {
                    synchronized (Ticker.this) {
                        Ticker.this.notify();
                    }
                    while (true) {
                        try {
                            updateThread.wait();
                        } catch (InterruptedException ex) {
                            Logger.getLogger(Ticker.class.getName()).log(Level.WARNING, null, ex);
                            continue;
                        }
                        try {
                            updateLoop();
                        } catch (InterruptedException ex) {
                            Logger.getLogger(Ticker.class.getName()).log(Level.FINE, null, ex);
                        }
                        Thread.interrupted();
                        // TODO: handle updateLoop returning without interrupt
                        synchronized (Ticker.this) {
                            image = null;
                            device = null;
                            dimension = null;
                        }
                    }
                }
            }
        });
        updateThread.setDaemon(true);
        updateThread.setName("Ticker" + id + "(" + name + "): updateThread");
        backgroundThread = new Thread(new Runnable() {

            public void run() {
                synchronized (backgroundThread) {
                    synchronized (Ticker.this) {
                        Ticker.this.notify();
                    }
                    backgroundLoop();
                }
            }
        });
        backgroundThread.setDaemon(true);
        backgroundThread.setName("Ticker" + id + "(" + name + "): backgroundThread");
        this.name = name;
    }

    /**
     * This method is called to start the renderThread, the updateThread and the
     * backgroundThread.
     */
    public final synchronized void start() {
        renderThread.start();
        try {
            wait();
        } catch (InterruptedException ex) {
            Logger.getLogger(Ticker.class.getName()).log(Level.SEVERE, null, ex);
        }
        updateThread.start();
        try {
            wait();
        } catch (InterruptedException ex) {
            Logger.getLogger(Ticker.class.getName()).log(Level.SEVERE, null, ex);
        }
        backgroundThread.start();
        try {
            wait();
        } catch (InterruptedException ex) {
            Logger.getLogger(Ticker.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Asynchronously activates this Ticker by notifying updateThread. isActive
     * will return true after this call returns until a call to deactivate has
     * returned.
     * @param device the device this Ticker will be activated on.
     */
    public final void activate(final GraphicsDevice device) {
        DisplayMode mode = device.getDisplayMode();
        activate(device, new Dimension(mode.getWidth(), mode.getHeight()));
    }

    /**
     * Asynchronously activates this Ticker by notifying updateThread. isActive
     * will return true after this call returns until a call to deactivate has
     * returned.
     * @param device the device this Ticker will be activated on.
     * @param dimension the size of the image that render should be producing.
     */
    public final void activate(final GraphicsDevice device, final Dimension dimension) {
        synchronized (this) {
            if (active) {
                return;
            }
            active = true;
            this.device = device;
            this.dimension = dimension;
        }
        synchronized (updateThread) {
            updateThread.notify();
        }
    }

    /**
     * Asynchronously deactivates this Ticker by interrupting updateThread.
     * isActive will return true after this call returns until a call to
     * activate has returned.
     */
    public final void deactivate() {
        synchronized (this) {
            if (!active) {
                return;
            }
            active = false;
        }
        updateThread.interrupt();
    }

    /**
     * Returns true iff this Ticker is active.
     */
    public final synchronized boolean isActive() {
        return active;
    }

    /**
     * Returns true iff this Ticker is active and is ready to be displayed. The
     * transition between one Ticker and another does not actually occur until
     * the next Ticker is ready.
     */
    public final synchronized boolean isReady() {
        return active && image != null;
    }

    /**
     * Check if the buffer has been updated since any previous call to isDirty.
     * Calling this function resets the dirty flag.
     * @return
     */
    public final boolean isDirty() {
        return dirty.compareAndSet(true, false);
    }

    /**
     * Draws the rendered image on to the specified Graphics.
     * @param g Graphics to paint on.
     */
    public synchronized final void paint(final Graphics g, final int width, final int height) {
        if (image != null) {
            if (image.getWidth(null) == width && image.getHeight(null) == height) {
                g.drawImage(image, 0, 0, null);
            } else {
                g.drawImage(image, 0, 0, width, height, null);
            }
        } else {
            g.setColor(Color.RED);
            g.fillRect(0, 0, width, height);
            g.setColor(Color.BLACK);
            g.drawString("null", 0, 20);
        }
    }

    /**
     * Registers the specified listener.
     * @param listener the listener to register.
     */
    public final void addPriorityChangeListener(final PriorityChangeListener listener) {
        synchronized (listeners) {
            listeners.add(listener);
        }
    }

    /**
     * Unregisters the specified listener.
     * @param listener the listener to unregister.
     */
    public final void removePriorityChangeListener(final PriorityChangeListener listener) {
        synchronized (listeners) {
            listeners.remove(listener);
        }
    }

    /**
     * Accesses the current priority of the Ticker.
     * @return the current priority.
     */
    public final synchronized PriorityEnum getPriority() {
        return priority;
    }

    /**
     * Accesses the current GraphicsDevice being rendered to by this Ticker, or
     * null if this Ticker is not active.
     * @return device.
     */
    protected final synchronized GraphicsDevice getDevice() {
        return device;
    }

    /**
     * Accesses the current ideal size of the image being rendered by render.
     * @return dimension.
     */
    protected final synchronized Dimension getDimension() {
        return (Dimension) dimension.clone();
    }

    /**
     * @return the name
     */
    public String getName() {
        return name;
    }

    /**
     * Accesses the current priority of the Ticker. Sends a PriorityChangeEvent
     * to registered listeners at some later time. This means that there is a
     * short period after calling this function that it still
     * @param priority
     */
    protected final void setPriority(final PriorityEnum priority) {
        final PriorityChangeEvent event;
        synchronized (this) {
            if (this.priority.equals(priority)) {
                return;
            }
            event = new PriorityChangeEvent(this, this.priority, priority);
            this.priority = priority;
        }
        synchronized (listeners) {
            for (final PriorityChangeListener e : listeners) {
//                SwingUtilities.invokeLater(new Runnable() {
//
//                    public void run() {
                e.priorityChanged(event);
//                    }
//                });
            }
        }
    }

    /**
     * Compares this Ticker to another Ticker in terms of priority.
     * @param o the other ticker to compare with.
     * @return a negative value if this ticker has a lower priority, a positive
     * value if this ticker has a higher priority, and zero if this ticker has
     * equal priority to the other ticker.
     */
    public int compareTo(final Ticker other) {
        int otherOrdinal;
        Date otherLastShown;
        synchronized (other) {
            otherOrdinal = other.priority.ordinal();
            otherLastShown = other.lastShown;
        }
        synchronized (this) {
            // FIXME: check sanity of these conditions
            int ret = otherOrdinal - priority.ordinal();
            return ret == 0 ? lastShown.compareTo(otherLastShown) : ret;
        }
    }

    /**
     * Asynchronously invokes render on the renderThread. Waits for any existing
     * call to render to complete, then calls render from the renderThread.
     * Should only be called from the updateThread.
     * @throws InterruptedException if the updateThread was interrupted.
     */
    protected final void invokeRender() throws InterruptedException {
        synchronized (renderThread) {
            renderThread.notify();
        }
    }

    /**
     * Synchronously invokes render on the renderThread. Waits for any existing
     * call to render to complete, then calls render from the renderThread.
     * Waits until the call to render completes before returning. Should only be
     * called from the updateThread.
     * @throws InterruptedException if the updateThread was interrupted.
     */
    protected final void invokeRenderAndWait() throws InterruptedException {
        synchronized (renderThread) {
            renderThread.notify();
            renderThread.wait();
        }
    }

    /**
     * Runs in the background changing the priority of this Ticker
     * appropriately. Less complex implementations are free to return from this
     * function if they no longer need to change the priority of this Ticker.
     * Called from the backgroundThread.
     */
    protected abstract void backgroundLoop();

    /**
     * Called when this Ticker is activated. This method should set up any
     * variables required for rendering, and then repeatedly call invokeRender
     * or invokeRenderAndWait when necessary until interrupted this method
     * should not return until interrupted.
     */
    protected abstract void updateLoop() throws InterruptedException;

    /**
     * Check if updateLoop can be interrupted at the moment. This function
     * should return true most of the time unless it is currently an
     * inappropriate time to interrupt the current Ticker. This method will be
     * called outside of updateThread, so should be synchronised properly.
     * @return true iff updateLoop can be interrupted right now.
     */
    public abstract boolean canDeactivate();

    /**
     * Check if updateLoop wants to be interrupted right now. This function
     * should return false until the ticker has shown everything it wants to
     * display for a reasonable amount of time for onlookers to observe it. This
     * method will be called outside of updateThread, so should be synchronised
     * properly.
     * @return true iff updateLoop should be interrupted right now.
     */
    public abstract boolean shouldDeactivate();

    /**
     * Produce an image to display, the ideal size of the image can be obtained
     * through a call to getDimension, but the image will be scaled to fit if it
     * is not this size. This method is called from within renderThread.
     */
    protected abstract Image render();
}
