package au.asn.ucc.ticker;

import au.asn.ucc.ticker.frontend.TickerFrame;
import au.asn.ucc.ticker.impl.TickerEnum;
import au.asn.ucc.ticker.impl.SimpleImageTicker;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * $Id: TickerManager.java 38 2010-01-20 12:02:15Z rvvs89 $
 * @author rvvs89
 */
public final class TickerManager {

    private boolean started;
    private final Thread managerThread;
    private final PriorityQueue<Ticker> queue;
    private final List<TickerFrame> displays;
    private final PriorityChangeListener listener;

    public TickerManager() {


        started = false;
        listener = new PriorityChangeListener() {

            public void priorityChanged(final PriorityChangeEvent event) {
                Ticker source = (Ticker) event.getSource();
                if (event.getOldVal() != PriorityEnum.DISABLED) {
                    synchronized (TickerManager.this) {
                        queue.remove(source);
                    }
                }
                if (event.getNewVal() != PriorityEnum.DISABLED) {
                    synchronized (TickerManager.this) {
                        queue.offer(source);
                    }
                }
            }
        };
        managerThread = new Thread(new Runnable() {

            public void run() {
                TickerManager.this.run();
            }
        });
        managerThread.setName("TickerManager init thread");
        // set up devices
        {
            GraphicsEnvironment environment =
                    GraphicsEnvironment.getLocalGraphicsEnvironment();
            GraphicsDevice[] devices = environment.getScreenDevices();
            StringBuilder info = new StringBuilder("System has ");
            info.append(devices.length);
            info.append(" displays.");
            displays = new ArrayList<TickerFrame>(devices.length);
            for (int i = 0; i < devices.length; i++) {
                GraphicsDevice device = devices[i];
                info.append("\nDisplay ");
                info.append(i + 1);
                info.append(" (");
                info.append(device.getIDstring());
                info.append(") ");
                if (device.isFullScreenSupported()) {
                    info.append("using active rendering.");
                } else {
                    info.append("using passive rendering.");
                }
                Logger.getLogger(TickerManager.class.getName()).log(Level.INFO, info.toString());
                TickerFrame display = new TickerFrame(device, this);
                displays.add(display);
            }
        }
        // set up queue
        {
            TickerEnum[] tickers = TickerEnum.values();
            queue = new PriorityQueue<Ticker>(tickers.length + displays.size() * 2);
            for (TickerEnum ticker : tickers) {
                pushTicker(ticker.getTicker());
            }
        }
    }

    public synchronized void start() {
        if (!started) {
            managerThread.start();
            started = true;
        }
    }

    private void run() {
        for (int i = displays.size() * 2; i > 0; i--) {
            pushTicker(new SimpleImageTicker("placeholder" + i, TickerEnum.class.getResource("/placeholder.png"), PriorityEnum.BELOW_NORMAL, 3000));
        }

        for (final TickerFrame display : displays) {
            display.start();
        }
    }

    /**
     * Inserts a Ticker into the queue.
     * @param ticker the Ticker to insert.
     */
    public synchronized void push(final Ticker ticker) {
        if (ticker == null) {
            return;
        }
        queue.offer(ticker);
        ticker.addPriorityChangeListener(listener);
    }

    /**
     * Called from a TickerFrame when the current Ticker has finished. A valid
     * Ticker will always be returned.
     * @param current the current Ticker being displayed.
     * @return the next Ticker to display.
     */
    public synchronized Ticker select() {
        Ticker ret = queue.poll();
        if (ret == null) {
            throw new IllegalStateException("Ticker leak detected! Queue returned null.");
        } else {
            ret.removePriorityChangeListener(listener);
        }
        return ret;
    }

    /**
     * Called from a TickerFrame to check if a Ticker with a higher priority
     * than the current Ticker is available, but not to commit to showing a new
     * Ticker unless that is the case.
     * @param current the Ticker currently being shown.
     * @return null if there is no Ticker with higher priority, or a Ticker with
     * higher priority.
     */
    public synchronized Ticker tentativeSelect(final Ticker current) {
        Ticker next = queue.peek();
        if (next == null) {
            throw new IllegalStateException("Ticker leak detected! Queue returned null.");
        }
        // FIXME: check sanity of this condition
        if (current.getPriority().ordinal() < next.getPriority().ordinal()) {
            next.removePriorityChangeListener(listener);
            return select();
        } else {
            return null;
        }
    }

    /**
     * Main method, called to start the application.
     * @param args command line arguments.
     */
    public static void main(final String[] args) {
        new TickerManager().start();
    }

    /**
     * Appends a Ticker into the priority queue.
     * @param ticker the Ticker to append.
     */
    private void pushTicker(final Ticker ticker) {
        if (ticker.getPriority() != PriorityEnum.DISABLED) {
            synchronized (this) {
                queue.offer(ticker);
            }
        }
        ticker.addPriorityChangeListener(listener);
        ticker.start();
    }
}
