package au.asn.ucc.ticker.frontend;

import au.asn.ucc.ticker.Ticker;
import au.asn.ucc.ticker.TickerManager;
import java.awt.Color;
import java.awt.DisplayMode;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GraphicsDevice;
import java.awt.image.BufferStrategy;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JFrame;

/**
 *
 * $Id: TickerFrame.java 36 2009-12-30 11:32:21Z rvvs89 $
 * @author rvvs89
 */
public final class TickerFrame extends JFrame {

    private final boolean exclusive;
    private final Thread updateThread;
    /**
     * This thread prepares the next Ticker for rendering.
     */
    private static final int SLEEP_LENGTH = 60;
    private final Thread activationThread;
    private final GraphicsDevice device;
    private final TickerManager manager;
    private final AtomicBoolean prepared;
    private Ticker current;
    private Ticker next;

    /** Creates new form TickerFrame */
    public TickerFrame(final GraphicsDevice device, final TickerManager manager) {
        super();
        initComponents();
        this.device = device;
        this.manager = manager;
        exclusive = device.isFullScreenSupported();
        if (exclusive) {
            updateThread = new Thread(new Runnable() {

                BufferStrategy strategy;

                private void callRender() {
                    Graphics g = strategy.getDrawGraphics();
                    if (!strategy.contentsLost()) {
                        drawTicker(g);
                        strategy.show();
                        g.dispose();
                    }
                    try {
                        Thread.sleep(SLEEP_LENGTH);
                    } catch (InterruptedException ex) {
                        Logger.getLogger(TickerFrame.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }

                public void run() {
                    // TODO: implement exclusive graphics behaviour
                    try {
                        TickerFrame.this.device.setFullScreenWindow(TickerFrame.this);
                        createBufferStrategy(2);
                        strategy = getBufferStrategy();
                        next = manager.select();
                        synchronized (activationThread) {
                            activationThread.notify();
                        }
                        // wait for initial activation
                        while (!prepared.compareAndSet(true, false)) {
                            Graphics g = strategy.getDrawGraphics();
                            if (!strategy.contentsLost()) {
                                drawLoading(g);
                                strategy.show();
                                g.dispose();
                            }
                            try {
                                Thread.sleep(SLEEP_LENGTH);
                            } catch (InterruptedException ex) {
                                Logger.getLogger(TickerFrame.class.getName()).log(Level.SEVERE, null, ex);
                            }
                        }
                        current = next;
                        next = null;
                        for (;;) {
                            while (!current.canDeactivate()) {
                                callRender();
                            }
                            while (!current.shouldDeactivate()) {
                                next = manager.tentativeSelect(current);
                                if (next != null) {
                                    break;
                                }
                                callRender();
                            }
                            if (next == null) {
                                next = manager.select();
                                synchronized (activationThread) {
                                    activationThread.notify();
                                }
                            }
                            while (!prepared.compareAndSet(true, false)) {
                                callRender();
                            }
                            current.deactivate();
                            manager.push(current);
                            current = next;
                            next = null;
                        }
                    } finally {
                        TickerFrame.this.device.setFullScreenWindow(null);
                    }
                }
            });
            updateThread.setName("Display(" + device.getIDstring() + "): updateThread");
        } else {
            updateThread = null;
            // FIXME: updateThread should monitor the current ticker and trigger repaints and ticker changes
        }
        activationThread = new Thread(new Runnable() {
            // FIXME: is this thread actually necessary? is a deadlock likely?

            public void run() {
                synchronized (activationThread) {
                    synchronized (TickerFrame.this) {
                        TickerFrame.this.notify();
                    }
                    for (;;) {
                        try {
                            activationThread.wait();
                        } catch (InterruptedException ex) {
                            Logger.getLogger(TickerFrame.class.getName()).log(Level.SEVERE, null, ex);
                        }
                        next.activate(device);
                        while (!next.isReady()) {
                            try {
                                Thread.sleep(SLEEP_LENGTH);
                            } catch (InterruptedException ex) {
                                Logger.getLogger(TickerFrame.class.getName()).log(Level.SEVERE, null, ex);
                            }
                        }
                        prepared.set(true);
                    }
                }
            }
        });
        activationThread.setDaemon(true);
        activationThread.setName("Display(" + device.getIDstring() + "): activationThread");
        prepared = new AtomicBoolean(false);
        setTitle("UCCTicker on " + device.getIDstring());
    }

    public void start() {
        if (!EventQueue.isDispatchThread()) {
            EventQueue.invokeLater(new Runnable() {

                public void run() {
                    TickerFrame.this.start();
                }
            });
        } else {
            synchronized (this) {
                activationThread.start();
                try {
                    wait();
                } catch (InterruptedException ex) {
                    Logger.getLogger(TickerFrame.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
            if (!exclusive) {
                device.setFullScreenWindow(this);
            } else {
                updateThread.start();
            }
        }
    }

    @Override
    public void paint(final Graphics g) {
        if (exclusive) {
            return;
        }
        drawTicker(g);
        repaint();
    }

    private void drawTicker(final Graphics g) {
        DisplayMode tmp = device.getDisplayMode();
        current.paint(g, tmp.getWidth(), tmp.getHeight());
    }

    private void drawLoading(final Graphics g) {
        g.setColor(Color.WHITE);
        DisplayMode tmp = device.getDisplayMode();
        g.fillRect(0, 0, tmp.getWidth(), tmp.getHeight());
        g.setColor(Color.BLACK);
        Font tmp2 = g.getFont();
        g.setFont(tmp2.deriveFont(tmp2.getSize2D() + 20));
        g.drawString("Loading...", 30, 50);
    }

    /**
     * This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setResizable(false);
        setUndecorated(true);
        getContentPane().setLayout(null);

        pack();
    }// </editor-fold>//GEN-END:initComponents
    // Variables declaration - do not modify//GEN-BEGIN:variables
    // End of variables declaration//GEN-END:variables
    private static final long serialVersionUID = 5099601784922646423L;
}
