package au.asn.ucc.ticker.impl;

import au.asn.ucc.ticker.PriorityEnum;
import au.asn.ucc.ticker.Ticker;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.CellRendererPane;
import org.bff.javampd.MPD;
import org.bff.javampd.MPDPlayer;
import org.bff.javampd.events.PlayerChangeEvent;
import org.bff.javampd.events.PlayerChangeListener;
import org.bff.javampd.exception.MPDConnectionException;
import org.bff.javampd.exception.MPDException;

/**
 *
 * @author rvvs89
 */
public final class MPDTicker extends Ticker {

    private static final int SLEEP_LENGTH = 1000;
    private static final int MIN_DISPLAY_LENGTH = 2000;
    private static final int MAX_DISPLAY_LENGTH = 5500;
    private final MPD mpd;
    private CellRendererPane renderPane = new CellRendererPane();
    private MPDStateRenderer renderer = new MPDTickerPanel();
    private MPDPlayer player;
    private Date displayed = null;
    private Date finished = null;
    private BufferedImage buffer = null;

    public MPDTicker(String hostname) {
        super("mpd@" + hostname);
        try {
            mpd = new MPD(hostname);
        } catch (UnknownHostException ex) {
            Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
            throw new IllegalArgumentException(ex);
        } catch (MPDConnectionException ex) {
            Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
            throw new IllegalArgumentException(ex);
        }
    }

    public MPDTicker(String hostname, int port) {
        super("mpd@" + hostname);
        try {
            mpd = new MPD(hostname, port);
        } catch (UnknownHostException ex) {
            Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
            throw new IllegalArgumentException(ex);
        } catch (MPDConnectionException ex) {
            Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
            throw new IllegalArgumentException(ex);
        }
    }

    public MPDTicker(String hostname, int port, String password) {
        super("mpd@" + hostname);
        try {
            mpd = new MPD(hostname, port, password);
        } catch (UnknownHostException ex) {
            Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
            throw new IllegalArgumentException(ex);
        } catch (MPDConnectionException ex) {
            Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
            throw new IllegalArgumentException(ex);
        }
    }

    @Override
    protected void backgroundLoop() {
        final Object lock = new Object();
        final AtomicBoolean active = new AtomicBoolean();
        synchronized (lock) {
            player = mpd.getMPDPlayer();
            player.addPlayerChangeListener(new PlayerChangeListener() {

                public void playerChanged(PlayerChangeEvent pce) {
                    switch (pce.getId()) {
                        case PlayerChangeEvent.PLAYER_STARTED:
                            synchronized (lock) {
                                active.set(true);
                                lock.notify();
                            }
                            break;
                        case PlayerChangeEvent.PLAYER_STOPPED:
                            synchronized (lock) {
                                active.set(false);
                                lock.notify();
                            }
                            break;
                    }
                }
            });
            try {
                if (player.getStatus() != MPDPlayer.PlayerStatus.STATUS_STOPPED) {
                    setPriority(PriorityEnum.NORMAL);
                }
            } catch (MPDException ex) {
                Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
            }
            for (;;) {
                try {
                    lock.wait();
                } catch (InterruptedException ex) {
                    Logger.getLogger(MPDTicker.class.getName()).log(Level.SEVERE, null, ex);
                }
                setPriority(active.get() ? PriorityEnum.NORMAL : PriorityEnum.DISABLED);
            }
        }
    }

    @Override
    protected void updateLoop() throws InterruptedException {
        try {
            displayed = new Date();
            finished = new Date();
            displayed.setTime(finished.getTime() + MIN_DISPLAY_LENGTH);
            finished.setTime(finished.getTime() + MAX_DISPLAY_LENGTH);
            do {
                invokeRenderAndWait();
                Thread.sleep(SLEEP_LENGTH);
            } while (getPriority() != PriorityEnum.DISABLED && new Date().before(finished));
        } finally {
            displayed = finished = null;
        }
    }

    @Override
    public boolean canDeactivate() {
        return getPriority() == PriorityEnum.DISABLED || displayed == null || new Date().after(displayed);
    }

    @Override
    public boolean shouldDeactivate() {
        return getPriority() == PriorityEnum.DISABLED || finished == null || new Date().after(finished);
    }

    @Override
    protected Image render() {
        Component stamp = renderer.getMPDStateRendererComponent(mpd);
        Dimension dim = getDimension();
        stamp.setSize(dim.width, dim.height);
        if(buffer == null || buffer.getWidth() != dim.width || buffer.getHeight() != dim.height) {
            buffer = new BufferedImage(dim.width, dim.height, BufferedImage.TYPE_INT_RGB);
        }
        Graphics g = buffer.getGraphics();
        renderPane.paintComponent(g, stamp, null, 0, 0, dim.width, dim.height, true);
        g.dispose();
        return buffer;
    }

    public interface MPDStateRenderer {

        public Component getMPDStateRendererComponent(MPD mpd);
    }
}
